/*  ARM 2.0 (Reference) Logging Agent  */

/*  This program is intended to be linked with an application
 *  which is under development/test that has been instrumented
 *  with calls to the ARM 2.0 API for the purpose of helping
 *  identify syntactic and semantic errors in the use of the
 *  ARM 2.0 API by the instrumented program.
 *
 *  Calls to the ARM API are sequentially logged to a specified
 *  log file (See 'User customization' below.) along with the
 *  content of significant parameters and data buffers.  Detected
 *  errors are logged to 'stderr' and the log file with "ERROR"
 *  embedded in the text.  Messages logged with "INFO" embedded
 *  in the text identify questionable semantic behavior by the
 *  instrumented program which may be intentional and is permitted
 *  by the ARM 2.0 API semantics.
 *
 *  Semantics of data collection vendor ARM agents may be more
 *  lenient than those of this agent.  The goal of this implementation
 *  is that an instrumented application which runs without errors
 *  with this implementation should normally run without errors with
 *  a vendor supplied agent.
 *
 *  This program is made available on an "AS IS" basis with no
 *  warranties made or implied as to its correctness or suitability
 *  for any purpose.
 *
 *  Author: Ken Zink
 *          Scientific and Engineering Software, Inc. (SES Inc.)
 *          September 1997
 */

#include "armagent.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


/*******  User customization  *******/

/*  log_name
 *
 *  Set the following string to the name of the file to which the ARM
 *  log data should be written.
 */

char log_name[] = "logfile";


/*******  Data definitions  *******/

FILE*   lf = 0;         /*  Log file descriptor  */

char    curr_time[] = "hh:mm:ss.sss:"; /*  Buffer for printable current time  */


/*  Structure definition for specification of meta_data for user-defined
 *  metrics for a transaction
 */

typedef struct arm_metadata_t {
    int32     type;             /*  Type of metric  */
    char*     name;             /*  Label for metric  */
} arm_metadata_t;


/*  Structure definition for metadata defined for a transaction
 */

typedef struct arm_meta101_t {
    arm_metadata_t metadata[7]; /*  Array of meta-data specs for this trans */
} arm_meta101_t;


/*  Structure definition for a transaction class member
 */

typedef struct arm_tran_t {
    int32     tran_id;          /*  Assigned tran_id  */
    arm_meta101_t* meta;        /*  Pointer to meta-data array  */
    struct arm_tran_t* next;    /*  Link to next transaction definition  */
} arm_tran_t;


/*  Structure definition for a transaction class defined by a user
 */

typedef struct arm_tran_class_t {
    char*     tran_name;        /*  Transaction name  */
    char*     tran_detail;      /*  Transaction detail  */
    arm_tran_t* tran_list;      /*  Transaction list  */
    struct arm_tran_class_t* next; /*  Link to next transaction class  */
} arm_tran_class_t;


/*  Structure definition for application users
 */

typedef struct arm_user_t {
    int32     appl_id;          /*  Assigned appl_id  */
    char*     user_name;        /*  User name  */
    int32     init_count;       /*  Count of initialized users  */
    arm_tran_class_t* class_list;/*  List of transaction classes  */
    struct arm_appl_t* appl;    /*  Application  */
    struct arm_user_t* next;    /*  Link to next user definition  */
    struct arm_user_t* prev;    /*  Link to previous user definition  */
} arm_user_t;


/*  Structure definition and head of list of defined applications
 */

typedef struct arm_appl_t {
    char*     appl_name;        /*  Application name  */
    arm_user_t* user_list;      /*  List of defined users  */
    struct arm_appl_t* next;    /*  Link to next application definition  */
    struct arm_appl_t* prev;    /*  Link to previous application definition  */
} arm_appl_t;

arm_appl_t*     appl_list = 0;  /*  List of defined applications  */


/*  Structure definition and head of list of active transactions
 */

typedef struct arm_active_tran_t {
    int32     handle;           /*  Assigned start_handle  */
    arm_user_t* user;           /*  Owner user  */
    arm_tran_t* tran;           /*  Transaction metadata  */
    arm_tran_class_t* tclass;   /*  Transaction class  */
    struct arm_active_tran_t* next; /*  Link to next active transaction  */
    struct arm_active_tran_t* prev; /*  Link to previous active transaction  */
} arm_active_tran_t;

arm_active_tran_t* arm_active_tran_list = 0;    /*  List of active transactions  */


/*******  Utility function prototypes  *******/


arm_active_tran_t* arm_find_handle( char* caller, int32 start_handle );
arm_user_t* arm_find_user( char* caller, int32 appl_id );
void arm_log_metrics( char* caller, arm_appl_t* appl, arm_user_t* user,
                      arm_tran_class_t* tclass, arm_tran_t* tran,
                      int32 start_handle, char* data, int32 data_size,
                      int corr_flag );
void arm_log_correlator( char* caller, char* msg, arm_user_t* user,
                         arm_tran_class_t* tclass, int32 start_handle,
                         arm_correlator_t* correlator );
void arm_log_blob( char* blob, int32 len );
void arm_set_curr_time();


/***********************************
 *                                 *
 *******  ARM API functions  *******
 *                                 *
 ***********************************/


/*  arm_init  -  initialize application
 *
 *  Register an application/user tuple; return a locally unique id for
 *  the tuple.  If a registration for the tuple already exists, return
 *  the id for that registration.
 *  NOTE: Duplicate registration calls are not logged unless LOGDUP is
 *        defined.
 *
 *  Returns    appl_id        < 0 implies an error
 */

int32 ARM_API arm_init(
    char*      appl_name,       /* Application name  */
    char*      appl_user_id,    /* User name         */  
    int32      flags,  	        /* Reserved = 0      */  
    char*      data,            /* Reserved = NULL   */
    int32      data_size)       /* Reserved = 0      */  
{
    static int32    appl_id = 0;        /*  Application id counter  */

    char*       name_temp = "";
    arm_appl_t* appl_temp;
    arm_user_t* user_temp;

    arm_set_curr_time();

    if ( lf == 0) {
        if ( ( lf = fopen( log_name, "w" ) ) == NULL ) {
            fprintf( stderr, "*** ARM Agent: %s arm_init: ERROR: Can not open "
                             "file <%s> for writing\n",
                     curr_time, log_name );
            return -1;
        }
    }    

    if ( ( appl_name == 0 ) || ( strlen( appl_name ) == 0 ) ) {
        fprintf( stderr, "*** ARM Agent: %s arm_init: ERROR: Appl_name "
                         "parameter is required\n",
                 curr_time );
        fprintf( lf, "%s arm_init: ERROR: Appl_name parameter is required\n",
                 curr_time );
        return -1;
    }

    /*  Search for an existing Application/User definition  */

    if ( appl_user_id != 0 ) {
        name_temp = appl_user_id;
    }

    for ( appl_temp = appl_list;
          appl_temp;
          appl_temp = appl_temp->next ) {

        if ( strncmp( appl_temp->appl_name, appl_name, 127 ) == 0 ) {
            for ( user_temp = appl_temp->user_list;
                  user_temp;
                  user_temp = user_temp->next ) {

                if ( strncmp( user_temp->user_name, name_temp, 127 ) == 0 ){
#ifdef LOGDUP
                    fprintf( lf, "%s arm_init: Application <%s> "
                                         "User <%s> = Appl_id <%d>\n",
                                     curr_time, appl_temp->appl_name,
                                     user_temp->user_name, user_temp->appl_id );
#endif /* LOGDUP */
                    return user_temp->appl_id; /*  Found user  */
                }
            }    

            break;              /*  Exit - found application but not user  */
        }
    }    

    if ( appl_temp == 0 ) {

        /*  Create new application entry  */

        appl_temp = (arm_appl_t*)malloc( sizeof( arm_appl_t ) );
        if ( strlen( appl_name ) < 128 ) {
            appl_temp->appl_name = strdup( appl_name );
        }
        else {
            appl_temp->appl_name = (char*)malloc( 128 );
            strncpy( appl_temp->appl_name, appl_name, 127 );
            appl_temp->appl_name[127] = 0;
            fprintf( stderr, "*** ARM Agent: %s arm_init: ERROR: "
                             "Application <%s> name truncated to 127 "
                             "characters\n",
                     curr_time, appl_temp->appl_name );
            fprintf( lf, "%s arm_init: ERROR: Application <%s> name truncated "
                         "to 127 characters\n",
                     curr_time, appl_temp->appl_name );
        }

        appl_temp->user_list = 0;

        /*  Add to list of defined applications  */

        if ( appl_list != 0 ) {
            appl_list->prev = appl_temp;
        }
        appl_temp->next = appl_list;
        appl_temp->prev = 0;
        appl_list = appl_temp;
    }

    /*  Create new user entry  */

    if ( ++appl_id < 0 ) appl_id = 1;           /*  Update and wrap to one  */

    user_temp = (arm_user_t*)malloc( sizeof( arm_user_t ) );

    if ( strlen( name_temp ) < 128 ) {
            user_temp->user_name = strdup( name_temp );
    }
    else {
        user_temp->user_name = (char*)malloc( 128 );
        strncpy( user_temp->user_name, name_temp, 127 );
        user_temp->user_name[127] = 0;
        fprintf( stderr, "*** ARM Agent: %s arm_init: ERROR: "
                         "Application <%s> User <%s> name truncated to 127 "
                         "characters\n",
                 curr_time, appl_temp->appl_name, user_temp->user_name );
        fprintf( lf, "%s arm_init: ERROR: Application <%s> User<%s> "
                     "name truncated to 127 characters\n",
                 curr_time, appl_temp->appl_name, user_temp->user_name );
    }

    user_temp->appl_id    = appl_id;
    user_temp->init_count = 1;
    user_temp->class_list = 0;
    user_temp->appl       = appl_temp;

    /*  Add new user to application user list  */

    if ( appl_temp->user_list != 0 ) {
        appl_temp->user_list->prev = user_temp;
    }
    user_temp->next = appl_temp->user_list;
    user_temp->prev = 0;
    appl_temp->user_list = user_temp;

    /*  Log call  */

    fprintf( lf, "%s arm_init: Application <%s> User <%s> = Appl_id <%d>\n",
             curr_time, appl_temp->appl_name, user_temp->user_name, appl_id );

    return appl_id;
}


/*  arm_getid  -  assign tranaction id
 *
 *  Register a transaction class for a user.  If that transaction
 *  has previously been registered for this user, the current
 *  transaction id (tran_id) is returned.
 *  NOTE: Duplicate transaction registrations are not logged unless
 *        LOGDUP is defined.
 *
 *  Returns    tran_id        < 0 implies an error
 */

int32 ARM_API arm_getid(
    int32      appl_id,        /* application handle            */
    char*      tran_name,      /* transaction name              */
    char*      tran_detail,    /* transaction additional info   */
    int32      flags,          /* Reserved = 0                  */
    char*      data,           /* format definition of user metrics */
    int32      data_size)      /* length of data buffer         */
{
    static int32    tran_id = 0;        /*  Transaction id counter  */

    int         i, f;
    int         num_fields = 0;
    char*       data_temp  = data;
    size_t      slen;
    arm_appl_t* appl_temp;
    arm_user_t* user_temp;
    arm_meta101_t*      meta_temp;
    arm_tran_class_t*   tclass_temp;
    arm_tran_t*         tran_temp;
    arm_user_data101_t* user_data;

    if ( appl_id < 0 )  return appl_id;         /*  arm_init failed  */

    arm_set_curr_time();

    if ( ( tran_name == 0 ) || ( ( slen = strlen( tran_name ) ) == 0 ) ){
        fprintf( stderr, "*** ARM Agent: %s arm_getid: ERROR: Tran_name "
                         "parameter is required\n",
                 curr_time );
        fprintf( lf, "%s arm_getid: ERROR: Tran_name parameter is required\n",
                 curr_time );

        return -1;
    }

    /*  Validate application id  */

    user_temp = arm_find_user( "arm_getid", appl_id );
    if ( user_temp == 0 ) return -1;

    appl_temp = user_temp->appl;

    /*  Check for prior transaction registration for this user  */

    user_data = (arm_user_data101_t*)data;

    if ( ( data_size < ( 8 + sizeof( arm_user_meta101_t ) ) ) ||
         ( user_data->format != ARM_Format101 ) ) {
        data_temp = 0;          /*  Assume no metric definitions  */
    }
    else {
        num_fields = (int)( ( data_size - 8 ) / sizeof( arm_user_meta101_t) );
    }

    for ( tclass_temp = user_temp->class_list;
          tclass_temp;
          tclass_temp = tclass_temp->next ) {

        /*  Compare transaction names  */

        if ( strncmp( tran_name, tclass_temp->tran_name, 127 ) == 0 ) break;
    }


    if ( tclass_temp ) {                /*  Found transaction class  */

        for ( tran_temp = tclass_temp->tran_list;
              tran_temp;
              tran_temp = tran_temp->next ) {

            /*  Compare meta-data definitions  */

            meta_temp = tran_temp->meta;

            if ( ( data_temp == 0 ) && ( meta_temp == 0 ) ) {
#ifdef LOGDUP
                fprintf( lf, "%s arm_getid: Application <%s> User <%s> "
                             "Transaction <%s> = Tran_id <%d>\n",
                         curr_time, appl_temp->appl_name, user_temp->user_name,
                         tclass_temp->tran_name, tran_temp->tran_id );
#endif /* LOGDUP */
                return tran_temp->tran_id; /*  Return previous registered id  */
            }

            if ( ( data_temp == 0 ) || ( meta_temp == 0 ) ) continue;

            for ( i = 0, f = ARM_Metric1_f; i < num_fields; ++i, f = f>>1 ) {
                if ( ( f & user_data->flags[1] ) != 0 ) {

                    if ( user_data->meta[i].type != meta_temp->metadata[i].type )
                        break;  /*  Metric types don't match  */

                    if ( strcmp( user_data->meta[i].name,
                                 meta_temp->metadata[i].name ) != 0 )
                        break;  /*  Metric names don't match  */
                }
                else {
                    if ( meta_temp->metadata[i].type != 0 ) break;
                }
            }    

            if ( i == num_fields ) {    /*  All passed metadata matched  */
                for ( ; i < 7; ++i ) {      /*  Check the rest  */
                    if ( meta_temp->metadata[i].type != 0 ) break;
                }
                if ( i == 7 ) {
#ifdef LOGDUP
                    fprintf( lf, "%s arm_getid: Application <%s> User <%s> "
                                 "Transaction <%s> = Tran_id <%d>\n",
                             curr_time, appl_temp->appl_name,
                             user_temp->user_name, tclass_temp->tran_name,
                             tran_temp->tran_id );
#endif /* LOGDUP */
                    return tran_temp->tran_id; /*  Return previous tran_id  */
                }
            }    
        }
    }    
    else {              /*  No existing transaction class  */

        tclass_temp = (arm_tran_class_t*)malloc( sizeof( arm_tran_class_t ) );

        if ( strlen( tran_name) < 128 ) {
            tclass_temp->tran_name = strdup( tran_name );
        }
        else {
            tclass_temp->tran_name = (char*)malloc( 128 );
            strncpy( tclass_temp->tran_name, tran_name, 127 );
            tclass_temp->tran_name[127] = 0;
            fprintf( stderr, "*** ARM Agent: %s arm_getid: ERROR: Application <%s> "
                             "User <%s> Transaction <%s> name truncated to 127 "
                             "characters\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     tclass_temp->tran_name );
            fprintf( lf, "%s arm_getid: ERROR: Application <%s> User<%s> "
                         "Transaction <%s> name truncated to 127 characters\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     tclass_temp->tran_name );
        }

        /*  Record any tran_detail  */

        if ( ( tran_detail != 0 ) && ( ( slen = strlen( tran_detail ) ) != 0 ) ) {
            if ( slen < 128 ) {
                tclass_temp->tran_detail = strdup( tran_detail );
            }
            else {
                tclass_temp->tran_detail = (char*)malloc( 128 );
                strncpy( tclass_temp->tran_detail, tran_detail, 127 );
                tclass_temp->tran_detail[ 127 ] = '\0';
                fprintf( stderr, "*** ARM Agent: %s arm_getid: ERROR: "
                                 "Application <%s> User <%s> Transaction <%s> "
                                 "Detail <%s> truncated to 127 characters\n",
                         curr_time, appl_temp->appl_name, user_temp->user_name,
                         tclass_temp->tran_name, tclass_temp->tran_detail );
                fprintf( lf, "%s arm_getid: ERROR: Application <%s> User <%s> "
                             "Transaction <%s> Detail <%s> truncated to 127 "
                             "characters\n",
                         curr_time, appl_temp->appl_name, user_temp->user_name,
                         tclass_temp->tran_name, tclass_temp->tran_detail );
            }

        }
        else {
            tclass_temp->tran_detail = strdup( "" );;
        }

        tclass_temp->tran_list = 0;

        /*  Push this transaction class on the user's list  */

        tclass_temp->next = user_temp->class_list;
        user_temp->class_list = tclass_temp;

        /*  Log transaction class  */

        fprintf( lf, "%s arm_getid: Application <%s> User <%s> "
                     "Transaction <%s> Detail <%s>\n",
                 curr_time, appl_temp->appl_name, user_temp->user_name,
                 tclass_temp->tran_name, tclass_temp->tran_detail );
    }

    /*  No match with prior registrations  */

    /***  Create transaction registration  ***/

    if ( ++tran_id < 0 ) tran_id  = 1;          /*  Update and wrap to one  */

    tran_temp = (arm_tran_t*)malloc( sizeof( arm_tran_t ) );

    tran_temp->tran_id = tran_id;

    /*  Log call  */

    fprintf( lf, "%s arm_getid: Application <%s> User <%s> Transaction <%s> = "
                 "Tran_id <%d>\n",
             curr_time, appl_temp->appl_name, user_temp->user_name,
             tclass_temp->tran_name, tran_temp->tran_id );



    /*  Save meta-data for user-defined metrics  */

    if ( ( data != 0 ) && ( data_size > 3 ) ) {
        if ( user_data->format == ARM_Format101 ) {

            num_fields = (int)(( data_size - 8 ) / sizeof( arm_user_meta101_t));

            if ( (int32)( ( num_fields * sizeof( arm_user_meta101_t) ) + 8 ) !=
                 data_size ) {
                fprintf( stderr, "*** ARM Agent: %s arm_getid: ERROR: "
                                 "Application <%s> User <%s> Transaction <%s> "
                                 "Data_size <%d> is not an even multiple of "
                                 "user metric definitions\n",
                         curr_time, appl_temp->appl_name, user_temp->user_name,
                         tclass_temp->tran_name, data_size );
                fprintf( lf, "%s arm_getid: ERROR: Application <%s> User <%s> "
                             "Transaction <%s> Data_size <%d> is not an even "
                             "multiple of user metric definitions\n",
                         curr_time, appl_temp->appl_name, user_temp->user_name,
                         tclass_temp->tran_name, data_size );
            }

            if ( num_fields > 0 ) {
                tran_temp->meta = (arm_meta101_t*)malloc( sizeof( arm_meta101_t) );
            }
            else {
                tran_temp->meta = 0;  /*  (Will fall through for loop)  */
            }

            meta_temp = tran_temp->meta;

            for ( i = 0, f = ARM_Metric1_f; i < num_fields; ++i, f = f>>1 ) {
                if ( ( f & user_data->flags[1] ) != 0 ) {

                    if ( ( user_data->meta[i].type > 0 ) &&
                         ( user_data->meta[i].type < ARM_MetricTypeLast ) ) {
                        meta_temp->metadata[i].type = user_data->meta[i].type;
                        slen = strlen( user_data->meta[i].name );
                        if ( slen < 44 ) {
                            meta_temp->metadata[i].name =
                                strdup( user_data->meta[i].name );
                        }
                        else {
                            meta_temp->metadata[i].name = (char*)malloc( 44 );
                            strncpy( meta_temp->metadata[i].name,
                                 user_data->meta[i].name, 43 );
                            meta_temp->metadata[i].name[43] = 0;

                            fprintf( stderr, "*** ARM Agent: %s arm_getid: "
                                     "ERROR: Application <%s> User <%s> "
                                     "Transaction <%s> Metric Field <%d> "
                                     "Name <%s> truncated to 43 characters\n",
                                 curr_time, appl_temp->appl_name,
                                 user_temp->user_name, tclass_temp->tran_name,
                                 (i+1), meta_temp->metadata[i].name );
                            fprintf( lf, "%s arm_getid: ERROR: Application <%s> "
                                     "User <%s> Transaction <%s> Metric Field <%d> "
                                     "Name <%s> truncated to 43 characters\n",
                                 curr_time, appl_temp->appl_name,
                                 user_temp->user_name, tclass_temp->tran_name,
                                 (i+1), meta_temp->metadata[i].name );
                        }


                        /*  Log metric definition  */

                        fprintf( lf, "%s arm_getid: Application <%s> User <%s> "
                                     "Transaction <%s> Metric Field <%d> "
                                     "Type <%d> Name <%s>\n",
                                 curr_time, appl_temp->appl_name,
                                 user_temp->user_name, tclass_temp->tran_name,
                                 (i+1), meta_temp->metadata[i].type,
                                 meta_temp->metadata[i].name );
                    }
                    else {
                        fprintf( stderr, "*** ARM Agent: %s arm_getid: "
                                         "ERROR: Application <%s> User <%s> "
                                         "Transaction <%s> Metric Field <%d> "
                                         "Type <%d> is not valid\n",
                                 curr_time, appl_temp->appl_name,
                                 user_temp->user_name, tclass_temp->tran_name,
                                 (i+1), user_data->meta[i].type );
                        fprintf( lf, "%s arm_getid: ERROR: Application <%s> "
                                     "User <%s> Transaction <%s> Metric "
                                     "Field <%d> Type <%d> is not valid\n",
                                 curr_time, appl_temp->appl_name,
                                 user_temp->user_name, tclass_temp->tran_name,
                                 (i+1), user_data->meta[i].type );
                    }
                }
                else {
                    meta_temp->metadata[i].type = 0;    /*  Clear undefined  */
                    meta_temp->metadata[i].name = 0;
                }
            }           /*  End for  */

            if ( num_fields > 0 ) {
                for ( i = num_fields; i < 7; ++i ) {
                    meta_temp->metadata[i].type = 0;    /*  Clear undefined  */
                    meta_temp->metadata[i].name = 0;
                }
            }    

        }           /*  End if ... format  */
        else {
            fprintf( stderr, "*** ARM Agent: %s arm_getid: ERROR: Application <%s> "
                             "User <%s> Transaction <%s> Data Format <%d> is "
                             "not valid\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     tclass_temp->tran_name, user_data->format );
            fprintf( lf, "%s arm_getid: ERROR: Application <%s> User <%s> "
                         "Transaction <%s> Data Format <%d> is not valid\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     tclass_temp->tran_name, user_data->format );
        }
    }           /*  End if ... data  */
    else {
        tran_temp->meta = 0;
    }

    /*  Push this definition on the class' list  */

    tran_temp->next = tclass_temp->tran_list;
    tclass_temp->tran_list = tran_temp;

    return tran_id;
}


/*  arm_start  -  mark beginning of a unique transaction

    Mark the beginning of a unique transaction; returns a unique handle
    for this active transaction.

    Returns    start_handle   < 0 implies an error
 */

int32 ARM_API arm_start(
    int32      tran_id,         /* transaction name identifier   */
    int32      flags,           /* Reserved = 0                  */
    char*      data,            /* user metrics data             */
    int32      data_size)       /* length of data buffer         */
{
    static int32   start_handle = 0;    /*  Transaction handle counter  */

    arm_appl_t*     appl_temp = 0;
    arm_user_t*     user_temp = 0;
    arm_tran_t*     tran_temp = 0;
    arm_tran_class_t*   tclass_temp = 0;
    arm_active_tran_t*  active_temp = 0;
    arm_user_data1_t*   user_data;
    arm_correlator_t*   correlator;

    if ( tran_id < 0 ) return tran_id;

    /*  Find transaction registration  */

    for ( appl_temp = appl_list;
          appl_temp;
          appl_temp = appl_temp->next ) {

        for ( user_temp = appl_temp->user_list;
              user_temp;
              user_temp = user_temp->next ) {

            for ( tclass_temp = user_temp->class_list;
                  tclass_temp;
                  tclass_temp = tclass_temp->next ) {

                for ( tran_temp = tclass_temp->tran_list;
                      tran_temp;
                      tran_temp = tran_temp->next ) {

                    if ( tran_temp->tran_id == tran_id ) break;
                }

                if ( tran_temp != 0 ) break;    /*  Found match  */
            }

            if ( tran_temp != 0 ) break;        /*  Found match  */
        }

        if ( tran_temp != 0 ) break;    /*  Found match  */
    }

    arm_set_curr_time( );

    if ( appl_temp == 0 ) {
        fprintf( stderr, "*** ARM Agent: %s arm_start: ERROR: Tran_id <%d> "
                         "is not currently valid\n",
                 curr_time, tran_id );
        fprintf( lf, "%s arm_start: ERROR: Tran_id <%d> is not currently valid\n",
                 curr_time, tran_id );

        return -1;              /*  No match found for tran_id  */
    }

    if ( ++start_handle < 0 ) start_handle = 1; /*  Update and wrap to one  */

    active_temp = (arm_active_tran_t*)malloc( sizeof( arm_active_tran_t ) );
    active_temp->handle  = start_handle;
    active_temp->user    = user_temp;
    active_temp->tran    = tran_temp;
    active_temp->tclass  = tclass_temp;

    if ( arm_active_tran_list != 0 ) {
        arm_active_tran_list->prev = active_temp;
    }
    active_temp->next = arm_active_tran_list;
    active_temp->prev = 0;
    arm_active_tran_list = active_temp;

    fprintf( lf, "%s arm_start: Application <%s> User <%s> Transaction <%s> = "
                 "Start_handle <%d>\n",
             curr_time, appl_temp->appl_name, user_temp->user_name,
             tclass_temp->tran_name, start_handle );

    /*  Log any application metrics  */

    arm_log_metrics( "arm_start", appl_temp, user_temp, tclass_temp, tran_temp,
                     start_handle, data, data_size, 1 );

    /*  Check for request for correlator generation  */

    if ( ( data != 0 ) && ( data_size > 39 ) ) {
        user_data = (arm_user_data1_t*)data;
        if ( user_data->flags[0] & ARM_CorrReq_f ) {
            correlator = (arm_correlator_t*)(&user_data->correlator);
            /*  Set bogus format code  */
            correlator->length = 40;
            correlator->format = 1;
            correlator->flags = 0;
            correlator->addr_format = 1;
            correlator->vendor_id = 29999;
            correlator->agent_version = 201;
            correlator->agent_instance = 42;
            correlator->tran_id = tran_id;
            correlator->start_handle = start_handle;
            strncpy( correlator->address, "Basil's Bar, Table 3", 20 );

            user_data->flags[0] = ( user_data->flags[0] & ~ARM_CorrGen_f ) | ARM_CorrGen_f;

            arm_log_correlator( "arm_start", "Generated correlator", user_temp,
                                tclass_temp, start_handle, correlator );
        }
    }    

    return start_handle;
}


/*  arm_update  -  indicate transaction progress

    Optional call to indicate progress of a long transaction or to provide
    additional information about a transaction.

    Returns    error_status   < 0 implies an error
 */

int32 ARM_API arm_update(
    int32      start_handle,    /* unique transaction handle     */
    int32      flags,           /* Reserved = 0                  */
    char*      data,            /* user metrics data             */
    int32      data_size)       /* length of data buffer         */
{
    int32     info_len;
    arm_active_tran_t*  active_temp;
    arm_user_data2_t*   user_data;

    if ( start_handle < 0 ) return start_handle;

    arm_set_curr_time( );

    active_temp = arm_find_handle( "arm_update", start_handle );
    if ( active_temp == 0 ) return -1 ;

    if ( ( data == 0 ) || ( data_size < 4 ) ) {
        /*  No data or metrics to log  */
        fprintf( lf, "%s arm_update: Application <%s> User <%s> "
                     "Transaction <%s> Start_handle <%d>\n",
                 curr_time, active_temp->user->appl->appl_name,
                 active_temp->user->user_name, active_temp->tclass->tran_name,
                 start_handle );
        return 0;
    }

    user_data = (arm_user_data2_t*)data;

    switch ( user_data->format ) {
        case ARM_Format1:       /*  Log application metrics  */

            arm_log_metrics( "arm_update", active_temp->user->appl,
                             active_temp->user, active_temp->tclass,
                             active_temp->tran, start_handle, data, data_size,
                             0 );
            break;

        case ARM_Format2:       /*  Print info data to log file  */

            if ( data_size < 1025 ) {
                info_len = data_size - 4;
            }
            else {
                fprintf( stderr, "*** ARM Agent: %s arm_update: ERROR: Application <%s> "
                                 "User <%s> Transaction <%s> Start handle <%d> "
                                 "Info truncated to 1020 characters\n",
                         curr_time, active_temp->user->appl->appl_name,
                         active_temp->user->user_name,
                         active_temp->tclass->tran_name, start_handle );
                fprintf( lf, "%s arm_update: ERROR: Application <%s> User<%s> "
                             "Transaction <%s> Start handle <%d> "
                             "Info truncated to 1020 characters\n",
                         curr_time, active_temp->user->appl->appl_name,
                         active_temp->user->user_name,
                         active_temp->tclass->tran_name, start_handle );

                info_len = 1020;
            }

            fprintf( lf, "%s arm_update: Application <%s> User <%s> "
                         "Transaction <%s> Start_handle <%d> Info <",
                     curr_time, active_temp->user->appl->appl_name,
                     active_temp->user->user_name,
                     active_temp->tclass->tran_name, start_handle );
            arm_log_blob( user_data->string1020, info_len );
            fprintf( lf, ">\n" );
            break;

        default:
            fprintf( stderr, "*** ARM Agent: %s arm_update: ERROR: Application <%s> "
                             "User <%s> Transaction <%s> Start_handle <%d> "
                             "Data format id <%d> is not valid\n",
                     curr_time, active_temp->user->appl->appl_name,
                     active_temp->user->user_name,
                     active_temp->tclass->tran_name, start_handle,
                     user_data->format );
            fprintf( lf, "%s arm_update: ERROR: Application <%s> User <%s> "
                         "Transaction <%s> Start_handle <%d> Data format "
                         "id <%d> is not valid\n",
                     curr_time, active_temp->user->appl->appl_name,
                     active_temp->user->user_name,
                     active_temp->tclass->tran_name, start_handle,
                     user_data->format );

            return -1;          /*  break;  */
    }

    return 0;
}


/*  arm_stop  -  mark end of a unique transaction

    Mark the completion or end of a unique transaction.

    Returns    error_status   < 0 implies an error
 */

int32 ARM_API arm_stop(
    int32      start_handle,   /* unique transaction handle     */
    int32      tran_status,    /* Good=0, Abort=1, Failed=2     */
    int32      flags,          /* Reserved = 0                  */
    char*      data,           /* user metrics data             */
    int32      data_size)      /* length of data buffer         */
{
    arm_active_tran_t* active_temp;

    if ( start_handle < 0 ) return start_handle;

    arm_set_curr_time( );

    active_temp = arm_find_handle( "arm_stop", start_handle );
    if ( active_temp == 0 ) return -1 ;

    fprintf( lf, "%s arm_stop: Application <%s> User <%s> Transaction <%s> "
                 "Start_handle <%d> Status <%d>\n",
             curr_time, active_temp->user->appl->appl_name,
             active_temp->user->user_name, active_temp->tclass->tran_name,
             start_handle, tran_status );

    /*  Log application metrics  */

    arm_log_metrics( "arm_stop", active_temp->user->appl, active_temp->user,
                     active_temp->tclass, active_temp->tran, start_handle, data,
                     data_size, 0 );

    /*  Unlink from the list of active transactions  */

    if ( active_temp->prev == 0 ) {
        arm_active_tran_list = active_temp->next;
        if ( active_temp->next != 0 ) {
            active_temp->next->prev = 0;
        }
    }    
    else {
        active_temp->prev->next = active_temp->next;
        if ( active_temp->next != 0 ) {
            active_temp->next->prev = active_temp->prev;
        }
    }    

    free( active_temp );        /*  Free the memory  */

    return 0;
}


/*  arm_end  -  end of application

    Mark the termination of a user use of an application or the
    termination of the application.

    Returns    error_status   < 0 implies an error
 */

int32 ARM_API arm_end(
    int32      appl_id,         /* application id                */
    int32      flags,           /* Reserved = 0                  */
    char*      data,            /* Reserved = NULL               */
    int32      data_size)       /* Reserved = 0                  */
{
    int         i;
    arm_appl_t* appl_temp;
    arm_user_t* user_temp;
    arm_tran_t* tran_temp;
    arm_tran_t* tran_next;
    arm_tran_class_t*   tclass_temp;
    arm_tran_class_t*   tclass_next;
    arm_active_tran_t*  active_temp;

    if ( appl_id < 0 )  return appl_id;         /*  arm_init failed  */

    /*  Validate application id  */

    arm_set_curr_time();
    user_temp = arm_find_user( "arm_end", appl_id );
    if ( user_temp == 0 ) return -1;

    if ( --( user_temp->init_count ) ) return 0;

    appl_temp = user_temp->appl;

    fprintf( lf, "%s arm_end: Application <%s> User <%s> appl_id <%d>\n",
             curr_time, appl_temp->appl_name, user_temp->user_name, appl_id );

    /*  Check for dangling transactions  */

    for ( active_temp = arm_active_tran_list;
          active_temp;
          active_temp = active_temp->next ) {
        if ( active_temp->user == user_temp ) {
            fprintf( stderr, "*** ARM Agent: %s arm_end: INFO: Application <%s> "
                             "User <%s> Transaction <%s> Start_handle <%d> "
                             "is not complete (stopped)\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     active_temp->tclass->tran_name, active_temp->handle );
            fprintf( lf, "%s arm_end: INFO: Application <%s> User <%s> "
                         "Transaction <%s> Start_handle <%d> "
                         "is not complete (stopped)\n",
                     curr_time, appl_temp->appl_name, user_temp->user_name,
                     active_temp->tclass->tran_name, active_temp->handle );
        }
    }    

    /***  Delete application/user tuple registration  ***/

    /*  First delete user transaction registrations  */

    for ( tclass_temp = user_temp->class_list;
          tclass_temp;
          tclass_temp = tclass_next ) {

        for ( tran_temp = tclass_temp->tran_list;
              tran_temp;
              tran_temp = tran_next ) {

            /*  Delete any meta-data for this transaction  */

            if ( tran_temp->meta ) {
                for ( i = 0; i < 7; ++i ) {
                    if ( tran_temp->meta->metadata[i].name != 0 ) {
                        free( tran_temp->meta->metadata[i].name );
                    }
                }

                free( tran_temp->meta );
            }

            tran_next = tran_temp->next;

            free( tran_temp );
        }

        tclass_next = tclass_temp->next;

        free( tclass_temp->tran_name );
        free( tclass_temp->tran_detail );

        free( tclass_temp );
    }

    /*  Delink user from apllication's user list  */

    if ( user_temp->prev == 0 ) {
        appl_temp->user_list = user_temp->next;
        if ( user_temp->next != 0 ) {
            user_temp->next->prev = 0;
        }
    }    
    else {
        user_temp->prev->next = user_temp->next;
        if ( user_temp->next != 0 ) {
            user_temp->next->prev = user_temp->prev;
        }
    }    

    free( user_temp );          /*  Free user entry  */

    /*  Check if all user registrations ended for this application  */

    if ( appl_temp->user_list == 0 ) {

        /*  Delink this application from the application list  */

        if ( appl_temp->prev == 0 ) {
            appl_list = appl_temp->next;
            if ( appl_temp->next != 0 ) {
                appl_temp->next->prev = 0;
            }
        }
        else {
            appl_temp->prev->next = appl_temp->next;
            if ( appl_temp->next != 0 ) {
                appl_temp->next->prev = appl_temp->prev;
            }
        }

        free( appl_temp->appl_name );
        free( appl_temp );      /*  Free application entry  */
    }

    return 0;
}


/***********************************
 *                                 *
 *******  Utility functions  *******
 *                                 *
 ***********************************/


/*  arm_find_handle  -  locate active transaction
 *
 *  Search the active transaction list for the given start_handle.
 *  Issue error messages if the start_handle is not a currently active
 *  transaction.
 *
 *  Returns    pointer to active transaction structure or zero
 */

arm_active_tran_t* arm_find_handle(
    char*     caller,
    int32     start_handle )
{
    arm_active_tran_t* active_temp;

    for ( active_temp = arm_active_tran_list;
          active_temp;
          active_temp = active_temp->next ) {

        if ( active_temp->handle == start_handle ) {

            return active_temp;         /*  Found match  */
        }
    }    

    fprintf( stderr, "*** ARM Agent: %s %s: ERROR: Start_handle <%d> is not a currently "
                     "active transaction\n",
             curr_time, caller, start_handle );
    fprintf( lf, "%s %s: ERROR: Start_handle <%d> is not a currently active transaction\n",
             curr_time, caller, start_handle );

    return active_temp;         /*  Return zero  */
}


/*  arm_find_user  -  locate application/user registration
 *
 *  Search the registered application/user tuples for the given appl_id.
 *  Issue error messages if the appl_id is not for a currently valid
 *  registration.
 *
 *  Returns    pointer to user structure or zero
 */

arm_user_t* arm_find_user(
    char*     caller,           /*  caller name  */
    int32     appl_id )         /*  application structure  */
{
    arm_appl_t* appl_temp;
    arm_user_t* user_temp;

    for ( appl_temp = appl_list;
          appl_temp;
          appl_temp = appl_temp->next ) {

        for ( user_temp = appl_temp->user_list;
              user_temp;
              user_temp = user_temp->next) {

            if ( user_temp->appl_id == appl_id ) {

                return user_temp;       /*  Found match  */
            }
        }
    }    

    fprintf( stderr, "*** ARM Agent: %s %s: ERROR: Appl_id <%d> is not currently valid\n",
             curr_time, caller, appl_id );
    fprintf( lf, "%s %s: ERROR: Appl_id <%d> is not currently valid\n",
             curr_time, caller, appl_id );

    return 0;
}


/*  arm_log_metrics  -  log user-defined metric data
 *
 *  Print user-defined metric data into the log file.
 *
 *  Returns    nothing (void)
 */

void arm_log_metrics(
    char*   caller,             /*  caller name  */
    arm_appl_t* appl,           /*  application  */
    arm_user_t* user,           /*  user  */
    arm_tran_class_t* tclass,   /*  transaction class  */
    arm_tran_t* tran,           /*  transaction  */
    int32   start_handle,       /*  start_handle  */
    char*   data,               /*  user data array  */
    int32   data_size,          /*  user data length  */
    int     corr_flag )         /*  log correlator flag  */
{
    int     i, f;
    int     loc_metrics, size_metric, str_metric;
    int     num_fields;
    int     slen;
    double  dtemp;
    arm_meta101_t*      meta_temp = tran->meta;
    arm_user_data1_t*   user_data;

    if ( meta_temp == 0 ) return;       /*  No meta definitions  */

    /*  Check for valid data field with correct format  */

    if ( ( data != 0 ) && ( data_size > 3 ) ) {
        user_data = (arm_user_data1_t*)data;
        if ( user_data->format == ARM_Format1 ) {

            loc_metrics = sizeof( user_data->format ) +
                          sizeof( user_data->flags ) +
                          sizeof( user_data->correlator);
            if ( ( data_size >= loc_metrics ) && corr_flag  &&
                 ( user_data->flags[0] & ARM_CorrPar_f ) ) {
                arm_log_correlator( caller, "Parent correlator", user, tclass,
                                    start_handle,
                                    (arm_correlator_t*)(&user_data->correlator) );
            }

            /*  Log metric fields that are defined and available  */

            size_metric = sizeof( arm_user_metric1_u );
            str_metric = loc_metrics + ( 6 * size_metric );
            if ( data_size >= loc_metrics + size_metric ) {
                if ( data_size >= str_metric ) {
                    num_fields = 6;
                }
                else {
                    num_fields = (int)(( data_size - loc_metrics ) / size_metric);
                    if ( ( ( num_fields * size_metric ) + loc_metrics ) != data_size ) {
                        fprintf( stderr, "*** ARM Agent: %s %s: ERROR: "
                                         "Application <%s> User <%s> "
                                         "Transaction <%s> Start_handle <%d> "
                                         "Data_size <%d> is not an even "
                                         "multiple of user metrics\n",
                                 curr_time, caller, appl->appl_name,
                                 user->user_name, tclass->tran_name,
                                 start_handle, data_size );
                        fprintf( lf, "*** ARM Agent: %s %s: ERROR: "
                                     "Application <%s> User <%s> Transaction <%s> "
                                     "Start_handle <%d> Data_size <%d> is not "
                                     "an even multiple of user metrics\n",
                                 curr_time, caller, appl->appl_name,
                                 user->user_name, tclass->tran_name,
                                 start_handle, data_size );
                    }
                }
                for ( i = 0, f = ARM_Metric1_f; i < num_fields; ++i, f = f>>1 ) {
                    if ( ( user_data->flags[1] & f ) == 0 ) continue;
                    if ( meta_temp->metadata[i].type == 0 ) continue;

                    if ( meta_temp->metadata[i].type >= ARM_MetricTypeLast ) {
                        fprintf( stderr, "%s %s: ERROR: Application <%s> User <%s> "
                                     "Transaction <%s> Start_handle <%d> "
                                     "Metric <%d> Type <%d> is not valid\n",
                                 curr_time, caller, appl->appl_name,
                                 user->user_name, tclass->tran_name,
                                 start_handle, (i+1), meta_temp->metadata[i].type );
                        fprintf( lf, "%s %s: ERROR: Application <%s> User <%s> "
                                     "Transaction <%s> Start_handle <%d> "
                                     "Metric <%d> Type <%d> is not valid\n",
                                 curr_time, caller, appl->appl_name,
                                 user->user_name, tclass->tran_name,
                                 start_handle, (i+1), meta_temp->metadata[i].type );

                        continue;       /*  Try next metric  */
                    }

                    fprintf( lf, "%s %s: Application <%s> User <%s> "
                                 "Transaction <%s> Start_handle <%d> "
                                 "Metric <%44s> : ",
                             curr_time, caller, appl->appl_name,
                             user->user_name, tclass->tran_name,
                             start_handle, meta_temp->metadata[i].name );

                    switch ( meta_temp->metadata[i].type ) {
                        case ARM_Counter32:
                            fprintf( lf, "<%d>\n", user_data->metric[i].counter32 );
                            break;
                        case ARM_Counter64:
#ifdef INT64
                            fprintf( lf, "<%ld>\n",
                                     user_data->metric[i].counter64 );
#else
                            fprintf( lf, "<%lf>\n",
                                     ( ( (double)(user_data->metric[i].counter64.upper ) * 4294967296.0 ) +
                                       (double)user_data->metric[i].counter64.lower ) );
#endif /* INT64 */
                            break;
                        case ARM_CntrDivr32:
                            dtemp = user_data->metric[i].cntrdivr32.count / user_data->metric[i].cntrdivr32.divisor;
                            fprintf( lf, "%d / %d = <%.4f>\n",
                                     user_data->metric[i].cntrdivr32.count,
                                     user_data->metric[i].cntrdivr32.divisor,
                                     dtemp );
                            break;
                        case ARM_Gauge32:
                            fprintf( lf, "<%d>\n", user_data->metric[i].gauge32 );
                            break;
                        case ARM_Gauge64:
#ifdef INT64
                            fprintf( lf, "<%ld>\n",
                                     user_data->metric[i].gauge64 );
#else
                            fprintf( lf, "<%lf>\n",
                                     ( ( (double)(user_data->metric[i].gauge64.upper ) * 2147483648.0 ) +
                                       (double)user_data->metric[i].gauge64.lower ) );
#endif /* INT64 */
                            break;
                        case ARM_GaugeDivr32:
                            dtemp = user_data->metric[i].gaugedivr32.gauge / user_data->metric[i].gaugedivr32.divisor;
                            fprintf( lf, "%d / %d = <%.4f>\n",
                                     user_data->metric[i].gaugedivr32.gauge,
                                     user_data->metric[i].gaugedivr32.divisor,
                                     dtemp );
                            break;
                        case ARM_NumericID32:
                            fprintf( lf, "<%u>\n",
                                     user_data->metric[i].numericid32 );
                            break;
                        case ARM_NumericID64:
#ifdef INT64
                            fprintf( lf, "<%lu>\n",
                                     user_data->metric[i].numericid64 );
#else
                            fprintf( lf, "<%lu>\n",
                                     user_data->metric[i].numericid64.upper );
                            fprintf( lf, "<%10lu>\n",
                                     user_data->metric[i].numericid64.lower );
#endif /* INT64 */
                            break;
                        case ARM_String8:
                            fprintf( lf, "<%8s>\n", user_data->metric[i].string8 );
                            break;
                        default:
                            break;
                    }
                }
            }    

            /*  Log String#1 metric if defined and available  */

            if ( ( data_size > 88 ) && ( ( user_data->flags[1] & ARM_String1_f ) != 0 ) &&
                 ( meta_temp->metadata[6].type != 0 ) ) {
                if ( data_size > 119 ) {
                    slen = 32;
                }
                else {
                    slen = (int)( data_size - 88 );
                }

                fprintf( lf, "%s %s: Metric <%44s> : <",
                         curr_time, caller, meta_temp->metadata[6].name );
                for ( i = 0; i < slen; ++i ) {
                    fprintf( lf, "%c", user_data->string32[i] );
                }
                fprintf( lf, ">\n" );
            }
        }
        else {
            fprintf( stderr, "*** ARM Agent: %s %s: ERROR: Application <%s> User <%s> "
                             "Transaction <%s> Data format id <%d> is not valid\n",
                     curr_time, caller, appl->appl_name, user->user_name,
                     tclass->tran_name, user_data->format );
            fprintf( lf, "%s $s: ERROR: Application <%s> User <%s> "
                         "Transaction <%s> Data format id <%d> is not valid\n",
                     curr_time, caller, appl->appl_name, user->user_name,
                     tclass->tran_name, user_data->format );
        }
    }           /*  End if ... data  */

    return;
}


/*  arm_log_correlator  -  log generated or user defined correlator
 *
 *  Print correlator in the log file.
 *
 *  Returns     nothing (void)
 */

void arm_log_correlator(
    char*     caller,           /*  calling function  */
    char*     msg,              /*  log message header  */
    arm_user_t* user,           /*  user  */
    arm_tran_class_t* tclass,   /*  transaction class  */
    int32     start_handle,     /*  transaction start_handle  */
    arm_correlator_t* correlator ) /*  correlator  */
{

    fprintf( lf, "%s %s: Application <%s> User <%s> Transaction <%s> "
                 "Start_handle <%d> %s: Length <%d> Format <%d> Flags <%x> "
                 "Address Format <%d> Vendor_id <%d> Agent_version <%d> "
                 "Agent Instance <%d> Start_handle <%d> Tran_id <%d> "
                 "Address <",
             curr_time, caller, user->appl->appl_name, user->user_name,
             tclass->tran_name, start_handle, msg, correlator->length,
             correlator->format, correlator->flags, correlator->addr_format,
             correlator->vendor_id, correlator->agent_version,
             correlator->agent_instance, correlator->start_handle,
             correlator->tran_id);
    arm_log_blob( correlator->address, (correlator->length - 20) );
    fprintf( lf, ">\n" );

    return;
}


/*  arm_log_blob  -  log an opaque blob in hex form
 *
 *  Print each of the characters of a blob into the log file in hex.
 *
 *  Returns     nothing (void)
 */

void arm_log_blob(
    char*     blob,             /*  blob array  */
    int32     len )             /*  length of blob array  */
{
    int  i;
    char c;

    for ( i = 0; i < len; ++i ) {
        c = blob[i];
        fprintf( lf, "%x", c );
    }

    return;
}


/*  arm_set_curr_time  -  acquire and format current (local) time
 *
 *  The current (local) time is acquired from the system and
 *  formatted into 'curr_time' with the following format:
 *     "hh:mm:ss.sss:"
 *
 *  Returns     nothing (void)
 */

void arm_set_curr_time()
{
    time_t t_time;
    char*  str;

    time( &t_time );
    str = ctime( &t_time );

    strncpy( curr_time, &(str[11]), 8 );

    return;
}


